/* ************************************************************************
 * arwparse.c
 *
 * Collection of function to parse ARW files
 *
 * Copyright 2010 Ronan Demoment.
 * ************************************************************************ */

/* ************************************************************************ */
/* Includes *************************************************************** */
/* ************************************************************************ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "arwparse.h"

/* ************************************************************************ */
/* Functions ************************************************************** */
/* ************************************************************************ */

/* -------------------------------------------------------------------------
 * This function loads the data from ARW file into memory and make basic
 * checks.
 *
 * tag_pointer pointer to the tag to parse
 *
 * Returns     0 if successful, 1 otherwise.
 * ------------------------------------------------------------------------- */
int load_arw(char *arwfilename, arwfile_descriptor *desc)
{
    FILE *arwfile;

    /* open ARW file */
    arwfile = fopen(arwfilename, "rb");
    if (arwfile == NULL)
    {
        perror(arwfilename);
        return(1);
    }

    /* retrieves file size */
    if (fseek(arwfile, 0, SEEK_END) != 0)
    {
        perror(arwfilename);
        fclose(arwfile);
        return(1);
    }
    desc->filesize = ftell(arwfile);

    /* rewind file */
    if (fseek(arwfile, 0, SEEK_SET) != 0)
    {
        perror(arwfilename);
        fclose(arwfile);
        return(1);
    }

    /* allocates memory for loading file */
    desc->data = (unsigned char *)malloc(desc->filesize);
    if (desc->data == NULL)
    {
        perror(arwfilename);
        fclose(arwfile);
        return(1);
    }

    /* loads file into memory */
    if (fread(desc->data, 1, desc->filesize, arwfile) != desc->filesize)
    {
        perror(arwfilename);
        fclose(arwfile);
        return(1);
    }

    /* initialize important fields of descriptor */
    strncpy(desc->filename, arwfilename, sizeof(desc->filename));
    desc->ifdcount = 0;

    /* close the file and exit successfully */
    fclose(arwfile);
    return(0);
}

/* -------------------------------------------------------------------------
 * Scan and parse all fields in a tag
 *
 * tag_pointer pointer to the tag to parse
 *
 * Returns     tag descriptor with all fields parsed
 * ------------------------------------------------------------------------- */
tag_descriptor parse_tag(unsigned char *tag_pointer)
{
    tag_descriptor tag;

    /* Get all fields of the tag */
    tag.pointer = tag_pointer;
    tag.id = *((short *)tag_pointer);
    tag.type = *((short *)(tag_pointer + 2));
    tag.count = *((long *)(tag_pointer + 4));
    tag.value = *((long *)(tag_pointer + 8));

    return(tag);
}

/* -------------------------------------------------------------------------
 * Scan and parse all tags in an IFD
 *
 * tiff_pointer pointer to the beginning of the tiff file
 * ifd_offset   offset to the IFD to parse
 *
 * Returns      IFD descriptor with all tags filled and parsed
 * ------------------------------------------------------------------------- */
ifd_descriptor parse_ifd(unsigned char *tiff_pointer,
                         unsigned long  ifd_offset)
{
    ifd_descriptor ifd;
    int            i;
    unsigned char  *tag_pointer;
    unsigned char  *ifd_pointer;

    ifd_pointer = tiff_pointer + ifd_offset;

    /* Reset subifd pointer */
    ifd.subifd = NULL;

    /* Get the tag count in the OFD */
    ifd.tagcount = *((unsigned short *)ifd_pointer);

    /* If tag count is more than 21, we are likely in front of an older A100 */
    /* file that are badly using subIFD.                                     */
    if (ifd.tagcount > 255) return(ifd);

    /* Get each tag from the IFD */
    tag_pointer = ifd_pointer + 2;
    for (i = 0; i < ifd.tagcount; i++)
    {
        ifd.tag[i] = parse_tag(tag_pointer);
        if ((ifd.tag[i].id == 0x14a) && (ifd.subifd == NULL))
        {
            ifd.subifd = (ifd_descriptor *)malloc(sizeof(ifd_descriptor));
            *ifd.subifd = parse_ifd(tiff_pointer, ifd.tag[i].value);
        }
        tag_pointer += 12;
        /* Get offset of the next IFD */
    }

    ifd.next_ifd = *((unsigned long *)tag_pointer);

    return(ifd);
}

/* -------------------------------------------------------------------------
 * Scan and parse all IFD in a ARW file
 *
 * desc     pointer to the arw file descriptor to be filled
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void parse_arw(arwfile_descriptor *desc)
{
    int ifd_offset;

    /* reset IFD counter */
    desc->ifdcount = 0;
    /* Get the first IFD offset from the header */
    ifd_offset = *((unsigned long *)(desc->data + 4));
    /* Parse as many IFD as we find */
    while (ifd_offset != 0)
    {
        /* Parse IFD */
        desc->ifd[desc->ifdcount] = parse_ifd(desc->data, ifd_offset);
        /* Get next IFD offset */
        ifd_offset = desc->ifd[desc->ifdcount].next_ifd;
        /* Update IFD count */
        desc->ifdcount++;
    };
}

/* -------------------------------------------------------------------------
 * Free all resources used by an ifd descriptor
 *
 * desc     pointer to the ifd descriptor to be purged
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void purge_ifd(ifd_descriptor *desc)
{
    if (desc->subifd != NULL)
    {
        purge_ifd(desc->subifd);
        free(desc->subifd);
    }
}

/* -------------------------------------------------------------------------
 * Free all resources used by an ARW descriptor
 *
 * desc     pointer to the arw file descriptor to be purged
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void purge_arw(arwfile_descriptor *desc)
{
    int i;

    for (i = 0; i < desc->ifdcount; i++)
    {
        purge_ifd(&desc->ifd[i]);
    }

    if (desc->data != NULL)
    {
        free(desc->data);
    }
}

/* -------------------------------------------------------------------------
 * Print data of a byte tag
 *
 * pointer  pointer to the byte array
 * length   number of bytes
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_byte_tag(unsigned char *pointer, int length)
{
    int i;

    for (i = 0; i < length; i++)
    {
        printf("0x%02x ", pointer[i]);
    }
}

/* -------------------------------------------------------------------------
 * Print data of a short tag
 *
 * pointer  pointer to the short array
 * length   number of shorts
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_short_tag(unsigned short *pointer, int length)
{
    int i;

    for (i = 0; i < length; i++)
    {
        printf("0x%04x (%d)", pointer[i], pointer[i]);
    }
}

/* -------------------------------------------------------------------------
 * Print data of a long tag
 *
 * pointer  pointer to the long array
 * length   number of longs
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_long_tag(unsigned long *pointer, int length)
{
    int i;

    for (i = 0; i < length; i++)
    {
        printf("0x%08lx (%ld)", pointer[i], pointer[i]);
    }
}

/* -------------------------------------------------------------------------
 * Print information about a tag entry
 *
 * arw_desc pointer to the arw file descriptor
 * ifd_desc pointer to the IFD descriptor
 * tag_desc pointer to the tag descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_tag_info(arwfile_descriptor *arw_desc, ifd_descriptor *ifd_desc, tag_descriptor *tag_desc)
{
    unsigned char *pointer;

    /* print ID and length */
    printf("  ID 0x%04x %5d items 0x%08x ", tag_desc->id, tag_desc->count, tag_desc->value);

    /* by default, data pointer is set to according to offset/value field */
    pointer = &arw_desc->data[tag_desc->value];

    /* check tag type */
    switch(tag_desc->type)
    {
        case 1:
            printf("BYTE      ");
            if (tag_desc->count <= 4)
                pointer = (unsigned char *)(&tag_desc->value);
            print_byte_tag(pointer, tag_desc->count);
            printf("\n");
            break;

        case 2:
            printf("ASCII     ");
            if (tag_desc->count <= 4)
                pointer = (unsigned char *)(&tag_desc->value);

            printf("%s\n", pointer);
            break;

        case 3:
            printf("SHORT     ");
            if (tag_desc->count <= 1)
                pointer = (unsigned char *)(&tag_desc->value);

            print_short_tag((unsigned short *)pointer, tag_desc->count);
            printf("\n");
            break;

        case 4:
            printf("LONG      ");
            if (tag_desc->count <= 1)
                pointer = (unsigned char *)(&tag_desc->value);

            print_long_tag((unsigned long *)pointer, tag_desc->count);
            printf("\n");
            break;

        case 5:
            printf("RATIONAL  ");

            printf("%ld / %ld\n", *((long *)pointer), *((long *)(pointer + 4)));
            break;

        case 7:
            printf("UNDEFINED ");
            if (tag_desc->count <= 4)
                pointer = (unsigned char *)(&tag_desc->value);

            printf("%s\n", pointer);
            break;

        default:
            printf("*** UNKNOWN (%d) ***\n", tag_desc->type);
            break;
    }
}

/* -------------------------------------------------------------------------
 * Print information about an IFD directory
 *
 * arw_desc pointer to the arw file descriptor
 * ifd_desc pointer to the IFD descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_ifd_info(arwfile_descriptor *arw_desc, ifd_descriptor *ifd_desc)
{
    int i;

    printf("  Next IFD : 0x%08x\n", ifd_desc->next_ifd);
    printf("  %d Tag entries\n", ifd_desc->tagcount);

    for (i = 0; i < ifd_desc->tagcount; i++)
    {
        print_tag_info(arw_desc, ifd_desc, &ifd_desc->tag[i]);
    }

    if (ifd_desc->subifd != NULL)
    {
        printf("*** SubIFD ***\n");
        print_ifd_info(arw_desc, ifd_desc->subifd);
        printf("\n");
    }
}

/* -------------------------------------------------------------------------
 * Print information about an ARW file.
 *
 * desc     pointer to the arw file descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_arw_info(arwfile_descriptor *desc)
{
    int i;

    printf("ARW filename : %s\n", desc->filename);
    printf("file size    : %ld bytes\n\n", desc->filesize);

    for (i =0; i < desc->ifdcount; i++)
    {
        printf("IFD #%d :\n", i);
        print_ifd_info(desc, &desc->ifd[i]);
    }
}

/* -------------------------------------------------------------------------
 * Find a specific tag in an IFD
 *
 * desc     pointer to the arw file descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
tag_descriptor *get_tag(ifd_descriptor *desc, const unsigned short ID)
{
    int             i;
    tag_descriptor *found_tag;

    found_tag = NULL;

    for (i = 0; i < desc->tagcount; i++)
    {
        if (desc->tag[i].id == ID)
            found_tag = &(desc->tag[i]);
    }

    return(found_tag);
}
