/* 
 * File:   arwparse.h
 * Author: Ronan Demoment
 *
 * Created on 14 février 2010, 13:29
 */

#ifndef _ARWPARSE_H
#define	_ARWPARSE_H

/* ************************************************************************ */
/* CONSTANTS ************************************************************** */
/* ************************************************************************ */

#define MAX_TAGS_IN_IFD 256

/* ************************************************************************ */
/* Types ****************************************************************** */
/* ************************************************************************ */

/* structure defining a TIFF tag */
typedef struct
{
    unsigned char  *pointer;
    unsigned short  id;
    unsigned short  type;
    unsigned int    count;
    unsigned int    value;
} tag_descriptor;

/* structure defining a TIFF directory */
typedef struct ifd_descriptor
{
    int             tagcount;
    tag_descriptor  tag[MAX_TAGS_IN_IFD];
    int             next_ifd;
    ifd_descriptor *subifd;
} ifd_descriptor;

/* structure gathering all data of an ARW file */
typedef struct
{
    char            filename[256];
    unsigned long   filesize;
    unsigned char  *data;
    int             ifdcount;
    ifd_descriptor  ifd[5];
} arwfile_descriptor;

/* ************************************************************************ */
/* Functions ************************************************************** */
/* ************************************************************************ */

/* -------------------------------------------------------------------------
 * This function loads the data from ARW file into memory and make basic
 * checks.
 *
 * tag_pointer pointer to the tag to parse
 *
 * Returns     0 if successful, 1 otherwise.
 * ------------------------------------------------------------------------- */
int load_arw(char *arwfilename, arwfile_descriptor *desc);

/* -------------------------------------------------------------------------
 * Scan and parse all fields in a tag
 *
 * tag_pointer pointer to the tag to parse
 *
 * Returns     tag descriptor with all fields parsed
 * ------------------------------------------------------------------------- */
tag_descriptor parse_tag(unsigned char *tag_pointer);

/* -------------------------------------------------------------------------
 * Scan and parse all tags in an IFD
 *
 * ifd_pointer pointer to the IFD to parse
 *
 * Returns     IFD descriptor with all tags filled and parsed
 * ------------------------------------------------------------------------- */
ifd_descriptor parse_ifd(unsigned char *ifd_pointer, unsigned long ifd_offset);

/* -------------------------------------------------------------------------
 * Scan and parse all IFD in a ARW file
 *
 * desc     pointer to the arw file descriptor to be filled
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void parse_arw(arwfile_descriptor *desc);

/* -------------------------------------------------------------------------
 * Free all resources used by an ARW descriptor
 *
 * desc     pointer to the arw file descriptor to be purged
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void purge_arw(arwfile_descriptor *desc);

/* -------------------------------------------------------------------------
 * Print data of a byte tag
 *
 * pointer  pointer to the byte array
 * length   number of bytes
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_byte_tag(unsigned char *pointer, int length);

/* -------------------------------------------------------------------------
 * Print data of a short tag
 *
 * pointer  pointer to the short array
 * length   number of shorts
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_short_tag(unsigned short *pointer, int length);

/* -------------------------------------------------------------------------
 * Print data of a long tag
 *
 * pointer  pointer to the long array
 * length   number of longs
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_long_tag(unsigned long *pointer, int length);

/* -------------------------------------------------------------------------
 * Print information about a tag entry
 *
 * arw_desc pointer to the arw file descriptor
 * ifd_desc pointer to the IFD descriptor
 * tag_desc pointer to the tag descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_tag_info(arwfile_descriptor *arw_desc, ifd_descriptor *ifd_desc, tag_descriptor *tag_desc);

/* -------------------------------------------------------------------------
 * Print information about an IFD directory
 *
 * arw_desc pointer to the arw file descriptor
 * ifd_desc pointer to the IFD descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_ifd_info(arwfile_descriptor *arw_desc, ifd_descriptor *ifd_desc);

/* -------------------------------------------------------------------------
 * Print information about an ARW file.
 *
 * desc     pointer to the arw file descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
void print_arw_info(arwfile_descriptor *desc);

/* -------------------------------------------------------------------------
 * Find a specific tag in an IFD
 *
 * desc     pointer to the arw file descriptor
 *
 * Returns  void
 * ------------------------------------------------------------------------- */
tag_descriptor *get_tag(ifd_descriptor *desc, const unsigned short ID);

#endif	/* _ARWPARSE_H */

